---------------------------------------------------------------------
--
-- Design unit:   Simple application module (reference design)
--
-- File name:     applic_wrp.vhd
--
-- Description:   Module containing 2 block of 128 Tero cells adn 2 counters to
--						count the number of oscillations of Tero cells,
--						One cell in each block are started together
--						with control signal and the number of oscillations
--						of the 2 cells are returned.
-- Parameters:
--
-- System:        VHDL'93
--
-- Autor:         Cédric Marchand, Hubert Curien Laboratory
--
-- Copyright:     LabHC 2013
--
-- Revision:      
--               Version 00, March 2013
--                - 
--
---------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_unsigned.all;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_arith.all;

LIBRARY WORK;
USE WORK.lab_hc_pkg.ALL;

ENTITY applic_wrp IS
  PORT(
    clk_ctrl      	: IN STD_LOGIC;           	-- Logic clock
    reset         	: IN STD_LOGIC;           	-- Resets control registers
-- Signals from controller module
	 enable_tero		: IN STD_LOGIC;			  	-- Signal to control oscillation start and stop
	 select_tero_1		: IN SLV_6_0;				  	-- Signal to select the Tero cell of the first block
	 select_tero_2		: IN SLV_6_0;					-- Signal to select the Tero cell of the second block
	 data_req			: IN STD_LOGIC;				-- Request for data

-- Signals to controller module
    appl_data        : OUT STD_LOGIC_VECTOR(31 downto 0) -- Signal containing datas
    );
END applic_wrp;

ARCHITECTURE rtl OF applic_wrp IS

-- Declaration of the Tero cell component : Hard Macro
COMPONENT tero_cycV is port(
	enable : in std_logic; -- Control signal
	output : out std_logic -- Output signal
);
end component;

---------------    Internal signals declaration    ------------------------
signal enable_1, enable_2 		: std_logic_vector(127 downto 0) := (others => '0'); -- Signals to control Tero cells oscillation
signal tero_out_1, tero_out_2 	: std_logic_vector(127 downto 0); -- Output signals
signal clk_count_1, clk_count_2	: std_logic; -- Clock of counter for number of Tero cells oscillations

signal data_gen_1, data_gen_2       : STD_LOGIC_VECTOR(15 DOWNTO 0) := (others => '0');  -- Signals for number of Tero cells oscillations

BEGIN

-- First Block of 128 Tero cells
tero_gen_1 : for i in 0 to 127 generate 
  tero_inst : tero_cycV port map(
		enable => enable_1(i), -- Control signal
		output => tero_out_1(i) -- Output signal
	);
end generate;

-- Second Block of 128 Tero cells
tero_gen_2 : for i in 0 to 127 generate 
	tero_inst : tero_cycV port map(
		enable => enable_2(i), -- Control signal
		output => tero_out_2(i) -- Output signal
	);
end generate;

-- Generation of control signals to control Tero cells oscillations
enable_1(conv_integer(select_tero_1)) <= enable_tero;
enable_2(conv_integer(select_tero_2)) <= enable_tero;
-- Generation of output signals used as clock to count the number of Tero cells oscillations
clk_count_1 <= tero_out_1(conv_integer(select_tero_1));
clk_count_2 <= tero_out_2(conv_integer(select_tero_2));

-- Process to count the number of oscillation of the Tero cell of the first block
PROCESS(reset, enable_1, tero_out_1, data_req, clk_count_1)
BEGIN
	IF (reset = '1' or data_req = '1') THEN -- reset or data request
		data_gen_1 <= (others => '0');
	ELSIF (rising_edge(clk_count_1) and enable_1(conv_integer(select_tero_1)) = '1' and data_gen_1<"1111111111111111") THEN
		data_gen_1 <= data_gen_1 + 1;
	END IF;
END PROCESS;

-- Process to count the number of oscillation of the Tero cell of the second block
PROCESS(reset, enable_2, tero_out_2, data_req, clk_count_2)
BEGIN
	IF (reset = '1' or data_req = '1') THEN -- reset or data request
		data_gen_2 <= (others => '0');
	ELSIF (rising_edge(clk_count_2) and enable_2(conv_integer(select_tero_2)) = '1' and data_gen_2<"1111111111111111") THEN
		data_gen_2 <= data_gen_2 + 1;
	END IF;
END PROCESS;

-- Generated data to be sent to the rest of the system
appl_data(15 downto 0) 	<= data_gen_1;
appl_data(31 downto 16) <= data_gen_2;

END rtl;