---------------------------------------------------------------------------
-- Design unit: Interface between cypress_if and the application module  --
---------------------------------------------------------------------------
-- Filename: sequencer.vhd                                               --
---------------------------------------------------------------------------
-- System: VHDL'93                                                       --
---------------------------------------------------------------------------
-- Authors:                                                              --
-- Ban J., Fischer V.                                                    --
-- (C) 2006, 2007 MICRONIC s.r.o.                                        --
---------------------------------------------------------------------------
-- Description:                                                          --
-- Data interface between Cypress EZ-USB2 interface FX2 and input/output --
-- FIFOs or depacketizer/packetizer. It controls data exchange protocol  --
-- and generates a timeout signal (pcktend).                             --
---------------------------------------------------------------------------
-- Parameters:                                                           --
--    * no parameters in this version                                    --
---------------------------------------------------------------------------
-- Inputs :                                                              --
--  * clk            : Cypress i/f clock                                 --
--  * reset          : global reset of the interface                     --
--  * dp_data        : FPGA input FIFO or depacketizer data              --
--  * dp_wr_req      : FPGA input FIFO or depacketizer write request     --
--  * p_rd_req       : FPGA output FIFO or packetizer read request       --
--  * data2bus       : 128-bit data from the application                 --
--  * state2bus      : 96-bit state from the application                 --
--  * busy2state     : application busy state bit (not used any more)    --
--                                                                       --
--                                                                       --
-- Outputs :                                                             --
--  * dp_full        : input FIFO in FPGA (or depacketizer) full flag    --
--                     active when FIFO (or depacketizer) is not ready   --
--  * p_data         : 16-bit data from FPGA output FIFO or packetizer   --
--  * p_empty        : output FIFO in FPGA (or packetizer) empty flag    --
--                     active when data from FIFO (or packetizer) are    --
--                     not available                                     --
--  * data2appl      : 128-bit data to the application                   --
--  * wr_data2appl   : 'write data to appl' flag (starts procedure)      --
--  * ctrl2appl      : 64-bit control word to the application            --
--  * wr_ctrl2appl   : writes control word to the application            --
--  * rd_data2bus    : reads data from the application                   --
--  * pckt_end       : signals packet end to the cypress if              --
--                                                                       --
---------------------------------------------------------------------------
-- Version: 2.00 July 2007                                               --
---------------------------------------------------------------------------
-- Revision history:                                                     --
-- Modifications in Version 2.00 (made by VF in July 2007)               --
--             - Modified file header                                    --
--             - Modified packet header - all headers are 128-bit long   --
--             - Byte counter replaced by a block counter (128-bit blcks)--
--                                                                       --
-- Remarks:   -------------------------------------------------------------
--  ......                                                               --
---------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.std_logic_unsigned.all;

ENTITY sequencer IS
  PORT (
    -- global
    clk           : IN STD_LOGIC;
    reset         : IN STD_LOGIC;
    mode_rst      : IN STD_LOGIC;
    
    -- cypress_if modul interface
    dp_data       : IN STD_LOGIC_VECTOR(15 DOWNTO 0);
    dp_wr_req     : IN STD_LOGIC;
    dp_full       : OUT STD_LOGIC;
    --
    p_data        : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
    p_rd_req      : IN STD_LOGIC;
    p_empty       : OUT STD_LOGIC;
    --
    pckt_end      : OUT STD_LOGIC;
    
  	-- application module interface
    data2appl     : OUT STD_LOGIC_VECTOR(127 DOWNTO 0); -- output to application module
    wr_data2appl  : OUT STD_LOGIC;                      -- data write strobe
    ctrl2appl     : OUT STD_LOGIC_VECTOR(63 DOWNTO 0);  -- output to command lines
    wr_ctrl2appl  : OUT STD_LOGIC;                      -- command write strobe
    --
    data2bus      : IN STD_LOGIC_VECTOR(127 DOWNTO 0);  -- input from application module
    rd_data2bus   : OUT STD_LOGIC;                      -- data read strobe
    state2bus     : IN STD_LOGIC_VECTOR(95 DOWNTO 0);   -- input from status lines

    trans_pending : OUT STD_LOGIC;                      -- a transaction is pending
    
    busy2state    : IN STD_LOGIC                        -- module is busy
  );
END sequencer;

ARCHITECTURE a OF sequencer IS

  -- depacketizer
  SIGNAL dp_waiting   : STD_LOGIC := '0'; -- the depacketizer has data waiting to be read
  SIGNAL dp_empty     : STD_LOGIC := '0'; -- depacketizer is empty
  SIGNAL dp_rd_req    : STD_LOGIC := '0'; -- depacketizer read request
  SIGNAL detect_com   : STD_LOGIC := '0'; -- command packet early detection
  SIGNAL detect_len   : STD_LOGIC := '0'; -- data packet header early detection
  SIGNAL detect_data  : STD_LOGIC := '0'; -- 128-bit data chuck early detection
  SIGNAL got_com      : STD_LOGIC := '0'; -- the early detection signals disappear once
  SIGNAL got_len      : STD_LOGIC := '0'; --   the depacketizer goes into wait mode, so
  SIGNAL got_data     : STD_LOGIC := '0'; --   these three signals save their values
  SIGNAL dp_valid     : STD_LOGIC_VECTOR(7 DOWNTO 0) := X"00"; -- to prevent detection errors
  SIGNAL dp_shift_reg : STD_LOGIC_VECTOR(127 DOWNTO 0):= X"00000000000000000000000000000000";
  SIGNAL ctrl2appl_i  : STD_LOGIC_VECTOR(63 DOWNTO 0);
  ALIAS word6         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(111 DOWNTO 96);
  ALIAS word5         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(95 DOWNTO 80);
  ALIAS word4         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(79 DOWNTO 64);
  ALIAS word3         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(63 DOWNTO 48);
  ALIAS word2         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(47 DOWNTO 32);
  ALIAS word1         : STD_LOGIC_VECTOR(15 DOWNTO 0) is dp_shift_reg(31 DOWNTO 16);
  SIGNAL pckt_end_i   : STD_LOGIC;        -- packet end internal flag
  SIGNAL terminate    : STD_LOGIC;        -- sends packet end
  SIGNAL read_status  : STD_LOGIC;        -- sends packet end

  -- sequencer
  TYPE state_type IS (sm_nothing,           -- wait for data from depacketizer
                      sm_write_command,     -- to application module
                      sm_wait2read_status,  -- wait for application module
                      sm_read_status,       -- read its status
                      sm_write_len,         -- write header of data packet
                      sm_wait2write_data,   -- wait for data chunk from depacketizer
                      sm_write_data,        -- write it to module
                      sm_wait2read_data,    -- wait for module to process it
                      sm_read_data,         -- read the result and send it to packetizer
                      sm_wait2just_read,    -- wait for module to generate data
                      sm_just_read);        -- read it and send it to packetizer
  SIGNAL sm_ctrl          : state_type;
  SIGNAL to_write_cnt     : STD_LOGIC_VECTOR(15 DOWNTO 0) := X"0000";
  SIGNAL to_read_cnt      : STD_LOGIC_VECTOR(15 DOWNTO 0) := X"0000";

  -- packetizer
  SIGNAL p_writing    : STD_LOGIC;  -- packetizer is writing data to cypress_if
  SIGNAL p_full       : STD_LOGIC;  -- same as above
  SIGNAL send_stat    : STD_LOGIC;  -- send a status packet
  SIGNAL send_head    : STD_LOGIC;  -- send a data packet header
  SIGNAL send_data    : STD_LOGIC;  -- send a data chunk
  SIGNAL p_header     : STD_LOGIC_VECTOR(127 DOWNTO 0); -- shift register for stat and len
  SIGNAL p_shift_reg  : STD_LOGIC_VECTOR(127 DOWNTO 0); -- shift register for data
  SIGNAL p_done       : STD_LOGIC_VECTOR(7 DOWNTO 0) := X"00"; -- shift reg used as a counter
  SIGNAL p_selector   : STD_LOGIC;  -- multiplexer selector
  SIGNAL blck_end     : STD_LOGIC;
--  SIGNAL p_last_data  : STD_LOGIC;
  

BEGIN
  trans_pending <= '0' WHEN ((sm_ctrl = sm_nothing) AND (p_writing = '0'))
                       ELSE '1';
                       
  ctrl2appl <= ctrl2appl_i;

  -- the depacketizer accepts 2 kinds of packets
  --
  -- packet         data format
  ------------------------------------------------------
  -- command        1337 c0de 0000 0000 CCCC CCCC CCCC CCCC
  -- data           1337 da7a 0000 0000 0000 0000 NNNN MMMM DDDD DDDD ...
  --
  -- where C...C = 64-bit control word
  --        NNNN = number of 128-bit blocks that will follow (D...D)
  --        MMMM = number of 128-bit blocks that are expected as an answer to this packet
  --
  -- packet      contents of dp_shift_reg when detection signals asserted
  -----------------------------------------------------------------------
  -- control         XXXX 1337 c0de 0000 0000 CCCC CCCC CCCC
  -- length          XXXX 1337 da7a 0000 0000 0000 0000 NNNN
  -- data            XXXX DDDD DDDD DDDD DDDD DDDD DDDD DDDD
  -- word index       7    6    5    4    3    2    1    0

  -- depacketizer detection signals
  detect_com <= '1' WHEN (word6 = X"1337") AND (word5 = X"c0de") AND
                       (dp_valid(6) = '1') AND (to_write_cnt = X"0000") ELSE '0';
  detect_len <= '1' WHEN (word6 = X"1337") AND (word5 = X"da7a") AND
                       (dp_valid(6) = '1') AND (to_write_cnt = X"0000") ELSE '0';
  detect_data <= '1' WHEN (dp_valid(6) = '1') AND (NOT (to_write_cnt = X"0000")) ELSE '0';

  -- depacketizer
  PROCESS(reset, clk) BEGIN
    IF reset = '1' THEN
      dp_waiting <= '0';
      dp_valid <= X"00";
    ELSIF rising_edge(clk) THEN
      -- data is flowing into the depacketizer
      IF dp_waiting = '0' THEN
        IF dp_wr_req = '1' THEN
          dp_shift_reg <= dp_shift_reg(111 DOWNTO 0) & dp_data;
          dp_valid     <= dp_valid(6 DOWNTO 0) & '1';
          got_com      <= detect_com;
          got_len      <= detect_len;
          got_data     <= detect_data;
          dp_waiting   <= detect_com or detect_len or detect_data;
        END IF;
      -- the depacketizer has to wait for the sequencer
      ELSE
        IF dp_rd_req = '1' THEN
          dp_valid   <= X"00";
          got_com    <= '0';
          got_len    <= '0';
          got_data   <= '0';
          dp_waiting <= '0';
        END IF;
      END IF;
    END IF;
  END PROCESS;

  -- output signals of depacketizer
  dp_full  <= dp_waiting;
  dp_empty <= NOT (dp_waiting);


  -- the sequencer has 2 counters:
  --
  -- to_write_cnt = # of 128-bit words to write to module
  -- to_read_cnt = # of 128-bit words to read from module
  --
  -- if (to_write_cnt > 0) and (to_read_cnt > 0), the wr_data2appl strobe is used to
  --   write data to the module and data is read out when busy2state goes low
  -- if (to_write_cnt = 0) and (to_read_cnt > 0), the rd_data2bus strobe is used to
  -- read data from the module (the outputs are assumed to be show-ahead)
  --
  -- the highest byte of ctrl2appl is held low when not written to
  -- (wr_ctrl2appl acts as an enable) - this is intended for dynamic inputs (reset...)
  -- for the other bytes, wr_ctrl2appl acts as a clock

  -- sequencer
  PROCESS(reset, clk) BEGIN
    IF reset = '1' THEN
      sm_ctrl <= sm_nothing;
      to_write_cnt <= X"0000";
      ctrl2appl_i <= (OTHERS => '0');     -- reset control word
--      data2appl   <= (OTHERS => '0');     -- reset data register      
    ELSIF rising_edge(clk) THEN
      case sm_ctrl is
        WHEN sm_nothing =>                  -- wait for data from depacketizer
          IF dp_empty = '0' THEN
            IF (got_com = '1') AND (busy2state = '0') THEN    -- command packet
              sm_ctrl     <= sm_write_command;
              ctrl2appl_i <= dp_shift_reg(63 DOWNTO 0);     -- store command
            ELSIF (got_len = '1') AND (p_full = '0') THEN -- data packet header
              sm_ctrl      <= sm_write_len;
              to_write_cnt <= dp_shift_reg(31 DOWNTO 16); -- store packet lengths
              to_read_cnt  <= dp_shift_reg(15 DOWNTO 0);
            END IF;
          END IF;

        WHEN sm_write_command =>            -- write command to module
          sm_ctrl <= sm_wait2read_status;
          ctrl2appl_i(63 DOWNTO 56) <= X"00"; -- reset dynamic outputs

        WHEN sm_wait2read_status =>         -- wait for module to process the command
          IF (busy2state = '0') AND (p_full = '0') THEN
            sm_ctrl <= sm_read_status;
          END IF;

        WHEN sm_read_status =>              -- write status to packetizer
         sm_ctrl <= sm_nothing;

        WHEN sm_write_len =>                -- write packet length word to packetizer
          IF to_write_cnt = X"0000" THEN
            IF to_read_cnt = X"0000" THEN
              sm_ctrl <= sm_nothing;          -- no data to write or read
            ELSE
              sm_ctrl <= sm_wait2just_read;   -- read first word
            END IF;
          ELSE
            sm_ctrl <= sm_wait2write_data;    -- write first word
          END IF;

        WHEN sm_wait2write_data =>          -- wait for data from depacketizer
          IF (dp_empty = '0') AND (got_data = '1') AND (busy2state = '0') THEN
            sm_ctrl <= sm_write_data;
--            data2appl <= dp_shift_reg;     -- store 128-bit data word
            to_write_cnt <= to_write_cnt - 1;   -- decrement write counter
          END IF;

        WHEN sm_write_data =>               -- write data to module
          IF to_read_cnt = X"0000" THEN
            IF to_write_cnt = X"0000" THEN
              sm_ctrl <= sm_nothing;          -- done with data packet
            ELSE
              sm_ctrl <= sm_wait2write_data;  -- just writing
            END IF;
          ELSE
            sm_ctrl <= sm_wait2read_data;     -- read the result
          END IF;

        WHEN sm_wait2read_data =>   -- wait for module and for packetizer to be ready
          IF (busy2state = '0') AND (p_full = '0') THEN
            sm_ctrl <= sm_read_data;
            to_read_cnt <= to_read_cnt - 1; -- decrement read counter
          END IF;

        WHEN sm_read_data =>                -- read data from module
          IF to_write_cnt = X"0000" THEN
            IF to_read_cnt = X"0000" THEN
              sm_ctrl <= sm_nothing;          -- done with data packet
            ELSE
              sm_ctrl <= sm_wait2just_read;   -- just reading
            END IF;
          ELSE
            sm_ctrl <= sm_wait2write_data;    -- write next word
          END IF;

        WHEN sm_wait2just_read =>   -- wait for module and for packetizer to be ready
          IF (busy2state = '0') AND (p_full = '0') THEN
            sm_ctrl <= sm_just_read;
            to_read_cnt <= to_read_cnt - 1;   -- decrement read counter
          END IF;

        WHEN sm_just_read =>                -- read data from module
          IF to_read_cnt = X"0000" THEN
              sm_ctrl <= sm_nothing;          -- done with data packet
          ELSE
            sm_ctrl <= sm_wait2just_read;     -- just reading
          END IF;

        WHEN OTHERS =>                      -- go to nothing
         sm_ctrl <= sm_nothing;

      END case;
    END IF;
  END PROCESS;

  PROCESS(reset, mode_rst, clk) BEGIN
    IF ((reset = '1') OR (mode_rst = '1')) THEN
      data2appl   <= (OTHERS => '0');       -- reset data register      
    ELSIF rising_edge(clk) THEN
      IF (sm_ctrl = sm_wait2write_data) THEN
        -- wait for data from depacketizer
        IF (dp_empty = '0') AND (got_data = '1') AND (busy2state = '0') THEN
          data2appl <= dp_shift_reg;        -- store 128-bit data word
        END IF;
      END IF;
    END IF;
  END PROCESS;
 
  -- sequencer output signals
  wr_ctrl2appl   <= '1' WHEN sm_ctrl = sm_write_command ELSE '0';
  wr_data2appl   <= '1' WHEN sm_ctrl = sm_write_data ELSE '0';
  rd_data2bus    <= '1' WHEN sm_ctrl = sm_just_read ELSE '0';
  
  -- signals to depacketizer
  dp_rd_req <= '1' WHEN (sm_ctrl = sm_write_command) or
                        (sm_ctrl = sm_write_len) or
                        (sm_ctrl = sm_write_data) ELSE '0';

  -- signals to packetizer
  send_stat <= '1' WHEN sm_ctrl = sm_read_status ELSE '0';
  send_head <= '1' WHEN (sm_ctrl = sm_write_len) AND (to_read_cnt /= X"0000") ELSE '0';
  send_data <= '1' WHEN (sm_ctrl = sm_read_data) or (sm_ctrl = sm_just_read) ELSE '0';

  -- the packetizer sends 2 types of packets:
  -- packet         data sent
  ----------------------------------------------------------
  -- status         1337 57a7 SSSS SSSS SSSS SSSS SSSS SSSS (128 bits)
  -- data           1337 da7a 0000 0000 0000 0000 NNNN 0000 DDDD DDDD ...
  --
  -- where: S...S = 96-bit status word
  --        NNNN  = number of 128-bit data blocks that will follow
  --     8x DDDD  = data in 128-bit blocks
  -- the data packet is sent by two signals - send_head sends the first 8 16-bit
  --   words (the header) and send_data sends a 128-bit data block

  -- packetizer
  PROCESS(reset, clk) BEGIN
    IF reset = '1' THEN
      p_writing <= '0';
    ELSIF rising_edge(clk) THEN
      blck_end   <= p_done(7);
      pckt_end_i <= p_done(7) AND (NOT blck_end) AND terminate AND read_status;
      -- packetizer is waiting for data from sequencer
      IF p_writing = '0' THEN
        IF send_stat = '1' THEN       -- status packet
          p_header   <= X"133757a7" & state2bus;
          p_done     <= "00000001";   -- write 8 words
          p_selector <= '1';
          p_writing  <= '1';
        ELSIF send_head = '1' THEN    -- length header
          p_header   <= X"1337da7a0000000000000000" & to_read_cnt & X"0000";
          p_done     <= "00000001";   -- write 8 words
          p_selector <= '1';
          p_writing  <= '1';
        ELSIF send_data = '1' THEN    -- data payload
          p_shift_reg <= data2bus;
          p_done      <= "00000001";  -- write 8 words
          p_selector  <= '0';
          p_writing   <= '1';
        END IF;
      -- packetizer is writing data to cypress_if
      ELSE
        IF p_rd_req = '1' THEN
          p_header(127 DOWNTO 16) <= p_header(111 DOWNTO 0);
          p_shift_reg(127 DOWNTO 16) <= p_shift_reg(111 DOWNTO 0);
          p_done(0) <= '0';
          p_done(7 DOWNTO 1) <= p_done(6 DOWNTO 0);
          p_writing <= NOT (p_done(7));  -- stop after '1' get shifted into msb
        END IF;
      END IF;
    END IF;
  END PROCESS;

  -- packetizer output signals
  p_data    <= p_header(127 DOWNTO 112) WHEN p_selector = '1' ELSE
               p_shift_reg(127 DOWNTO 112);
  p_full    <= p_writing;
  p_empty   <= NOT (p_writing);

  pckt_end  <= pckt_end_i;

  PROCESS(reset, clk, sm_ctrl)
  BEGIN
    IF reset = '1' THEN
      terminate   <= '0';
      read_status <= '0';
    ELSIF rising_edge(clk) THEN
      IF sm_ctrl = sm_write_command THEN
        terminate <= ctrl2appl_i(63);
      ELSIF pckt_end_i = '1' THEN
        terminate <= '0';
      END IF;
      --
      IF sm_ctrl = sm_read_status THEN
        read_status <= '1';
      ELSIF ((pckt_end_i = '1') OR (sm_ctrl = sm_write_command)) THEN
        read_status <= '0';
      END IF;
    END IF;
  END PROCESS;

END a;